'''Provides functions to parse the text-file logs produced by E-Prime experiments'''

def read_header(filename):
	''' returns a dictionary of values in the header of the given file '''
	header = {}
	in_header = False
	with open(filename,'rU') as f:
		lines = [x.strip() for x in f.read().split('\n')]
		for line in lines:
			if line=="*** Header Start ***":
				in_header=True
				continue
			if line=="*** Header End ***":
				return header
			fields = line.split(": ")
			if len(fields)==2:
				header[fields[0]] = fields[1]

def parse_frames(filename):
	''' quick and dirty eprime txt file parsing - doesn\'t account for nesting 
	
	**Example usage**::
	
		for frame in neural.eprime.parse_frames("experiment-1.txt"):
			trial_type = frame['TrialSlide.Tag']
			trial_rt = float(frame['TrialSlide.RT'])
			print '%s: %fms' % (trial_type,trial_rt)
	'''
	frames = []
	with open(filename,'rU') as f:
		frame = {}
		lines = [x.strip() for x in f.read().split('\n')]
		for line in lines:
			if line == '*** LogFrame Start ***':
				frame = {}
			if line == '*** LogFrame End ***':
				frames.append(frame)
				yield frame
			fields = line.split(": ")
			if len(fields)==2:
				frame[fields[0]] = fields[1]
	#return frames


'''NeurAL -- Neuroimaging Analysis Library

Library to provide helper functions to make analyzing
neuroimaging data in python a little easier'''

version = 0.48

#! flag to indicate whether most functions should be verbose or not
verbose = True

import neural
from notification import level,notify

from utils import *

import afni
import fsl
import eprime
import dicom
import notification


# functions that are useful/general enough to be on top-level, not in a specific module:
from afni import dset_info,subbrick,cdf,voxel_count,prefix,suffix



''' wrapper functions for common AFNI tasks '''
import re,subprocess,os,glob
import multiprocessing
import neural
import neural as nl
import platform
from operator import mul

def _open_X11_ports():
    tcp_ports = []
    if platform.system()=='Darwin':
        tcp_ports = subprocess.check_output('lsof -i -n | awk \'$9 ~ /:60[0-9][0-9]$/ {split($9,a,":"); print a[length(a)]}\' | sort | uniq',shell=True).strip().split('\n')
    if platform.system()=='Linux':
        tcp_ports = subprocess.check_output('netstat -ntlp | awk \'$6=="LISTEN" && $4 ~ /:60[0-9][0-9]$/ {split($4,a,":"); print a[length(a)]}\' | sort | uniq',shell=True).strip().split('\n')
    local_sockets = []
    if os.path.exists('/tmp/.X11-unix/'): 
        local_sockets = [x[1:] for x in os.listdir('/tmp/.X11-unix/')]
    return ['localhost:%d' % (int(x)-6000) for x in tcp_ports] + [':%s' % x for x in local_sockets]
    
def openX11(dsets=[]):
    ''' hack-y style method to try to open an AFNI window with the directory or datasets given 
    
    if this function is useful, it should be made more stable/portable'''
    my_env = os.environ.copy()
    if 'DISPLAY' not in my_env or my_env['DISPLAY']=='':
        open_ports = _open_X11_ports()
        if len(open_ports)==0:
            raise Exception('Couldn\'t find any open X11 ports')
        my_env['DISPLAY'] = open_ports[0]
    subprocess.Popen(['afni'] + dsets, env=my_env)

def _dset_raw_info(dset):
    ''' returns raw output from running ``3dinfo`` '''
    return subprocess.check_output(['3dinfo','-verb',str(dset)],stderr=subprocess.STDOUT)

class DsetInfo:
    ''' contains organized output from ``3dinfo`` 
    
    All lists are returned in RAI order (i.e., a list of 3 numbers refers to the R-L axis,
    A-P axis, then I-S axis)'''
    def __init__(self):
        self.subbricks = []
        self.voxel_size = []        #! size of voxel in mm, listed in LPI order
        self.voxel_volume = None    #! volume of voxel in mm^3
        self.voxel_dims = []
        self.spatial_from = []
        self.spatial_to = []
        self.slice_timing = None
        self.TR = None
        self.orient = None
    
    def subbrick_labeled(self,label):
        for i in xrange(len(self.subbricks)):
            if self.subbricks[i]['label']==label:
                return i
        raise LookupError

def dset_info(dset):
    ''' runs ``3dinfo`` and returns a :class:`DsetInfo` object containing the results '''
    info = DsetInfo()
    raw_info = _dset_raw_info(dset)
    # Subbrick info:
    sub_info = re.findall(r'At sub-brick #(\d+) \'([^\']+)\' datum type is (\w+).*(\n.*statcode = (\w+);  statpar = (.*)|)',raw_info)
    for brick in sub_info:
        info.subbricks.append({
            'label': brick[1],
            'datum': brick[2],
            'stat': brick[4],
            'params': brick[5].split()
        })
    info.reps = len(info.subbricks)
    # Dimensions:
    
    orient = re.search('\[-orient ([A-Z]+)\]',raw_info)
    if orient:
        info.orient = orient.group(1)
    for axis in ['RL','AP','IS']:
        m = re.search(r'%s-to-%s extent:\s+([0-9-.]+) \[%s\] -to-\s+([0-9-.]+) \[%s\] -step-\s+([0-9-.]+) mm \[([0-9]+) voxels\]' % (axis[0],axis[1],axis[0],axis[1]),raw_info)
        if m:
            info.spatial_from.append(float(m.group(1)))
            info.spatial_to.append(float(m.group(2)))
            info.voxel_size.append(float(m.group(3)))
            info.voxel_dims.append(float(m.group(4)))
    if len(info.voxel_size)==3:
        info.voxel_volume = reduce(mul,info.voxel_size)
    
    slice_timing = re.findall('-time:[tz][tz] \d+ \d+ [0-9.]+ (.*?) ',raw_info)
    if len(slice_timing):
        info.slice_timing = slice_timing[0]
    TR = re.findall('Time step = ([0-9.]+)s',raw_info)
    if len(TR):
        info.TR = float(TR[0])
    
    # Other info..
    details_regex = {
        'identifier': r'Identifier Code:\s+(.*)',
        'filetype': r'Storage Mode:\s+(.*)'
    }
    
    for d in details_regex:
        m = re.search(details_regex[d],raw_info)
        if m:
            setattr(info,d,m.group(1))
    
    return info

def subbrick(dset,label,coef=False,tstat=False,fstat=False,rstat=False):
    ''' returns a string referencing the given subbrick within a dset
    
    This method reads the header of the dataset ``dset``, finds the subbrick whose
    label matches ``label`` and returns a string of type ``dataset[X]``, which can
    be used by most AFNI programs to refer to a subbrick within a file
    
    The options coef, tstat, fstat, and rstat will add the suffix that is
    appended to the label by 3dDeconvolve
    
    :coef:  "#0_Coef"
    :tstat: "#0_Tstat"
    :fstat: "_Fstat"
    :rstat: "_R^2"
    '''
    
    if coef:
        label += "#0_Coef"
    elif tstat:
        label += "#0_Tstat"
    elif fstat:
        label += "_Fstat"
    elif rstat:
        label += "_R^2"
    
    info = dset_info(dset)
    i = info.subbrick_labeled(label)
    return '%s[%d]' % (dset,i)

def calc(dsets,expr,prefix=None,datum=None):
    ''' returns a string of an inline 3dcalc expression
    
    ``dsets`` can be a single string, or list of strings. Each string in ``dsets`` will
    be labeled 'a','b','c', sequentially. The expression ``expr`` is used directly
    
    If ``prefix`` is not given, will return a 3dcalc string that can be passed to another 
    AFNI program as a dataset. Otherwise, will create the dataset with the name ``prefix``'''
    if isinstance(dsets,basestring):
        dsets = [dsets]
    if prefix:
        cmd = ['3dcalc']
    else:
        cmd = ['3dcalc(']
    
    for i in xrange(len(dsets)):
        cmd += ['-%s'% chr(97+i),dsets[i]]
    cmd += ['-expr',expr]
    if datum:
        cmd += ['-datum',datum]
    
    if prefix:
        cmd += ['-prefix',prefix]
        return nl.run(cmd,products=prefix)
    else:
        cmd += [')']
        return ' '.join(cmd)

def cdf(dset,p,subbrick=0):
    ''' converts *p*-values to the appropriate statistic for the specified subbrick '''
    info = dset_info(dset)
    command = ['cdf','-p2t',info.subbricks[subbrick]['stat'],str(p)] + info.subbricks[subbrick]['params']
    return float(subprocess.check_output(command).split()[2])

def thresh_at(dset,p,subbrick=0,positive_only=False,output_suffix=None):
    ''' returns a string containing an inline ``3dcalc`` command that thresholds the 
        given dataset at the specified *p*-value, or will create a new dataset if a 
        suffix is given '''
    t = cdf(dset,p,subbrick)
    expr = 'astep(a,%f)' % t
    if positive_only:
        expr = 'abs(a)*step(a-%f)' % t
    subref = '-a%d' % subbrick
    if subbrick==0 and dset[-1]==']':
        subref = '-a'
    if output_suffix==None:
        return '3dcalc( %s %s -expr %s )' % (subref,dset,expr)
    neural.run(['3dcalc',subref,dset,'-expr',expr,'-prefix',suffix(dset,output_suffix)])

def cluster(dset,min_distance,min_cluster_size,prefix):
    ''' runs 3dmerge to cluster given dataset '''
    neural.run(['3dmerge','-1clust',min_distance,min_cluster_size,'-prefix',prefix,dset])

def voxel_count(dset,subbrick=0,p=None,positive_only=False,mask=None,ROI=None):
    ''' returns the number of non-zero voxels
    
    :subbrick:      use the given subbrick
    :p:             threshold the dataset at the given *p*-value, then count
    :positive_only: only count positive values
    :mask:          count within the given mask
    :ROI:           only use the ROI with the given value (or list of values) within the mask
                    if ROI is 'all' then return the voxel count of each ROI
                    as a dictionary
    '''
    if p:
        dset = thresh_at(dset,p,subbrick,positive_only)
    else:
        if not dset.startswith('3dcalc('):
            dset = '%s[%d]' % (dset,subbrick)
            if positive_only:
                dset = calc(dset,'step(a)')
    
    count = 0
    devnull = open(os.devnull,"w")
    if mask:
        cmd = ['3dROIstats','-1Dformat','-nomeanout','-nobriklab', '-nzvoxels']
        cmd += ['-mask',str(mask),str(dset)]
        out = subprocess.check_output(cmd,stderr=devnull).split('\n')
        if len(out)<4:
            return 0
        rois = [int(x.replace('NZcount_','')) for x in out[1].strip()[1:].split()]
        counts = [int(x.replace('NZcount_','')) for x in out[3].strip().split()]
        count_dict = None
        if ROI==None:
            ROI = rois
        if ROI=='all':
            count_dict = {}
            ROI = rois
        else:
            if not isinstance(ROI,list):
                ROI = [ROI]
        for r in ROI:
            if r in rois:
                roi_count = counts[rois.index(r)]
                if count_dict!=None:
                    count_dict[r] = roi_count
                else:
                    count += roi_count
    else:
        cmd = ['3dBrickStat', '-slow', '-count', '-non-zero', str(dset)]
        count = int(subprocess.check_output(cmd,stderr=devnull).strip())
    if count_dict:
        return count_dict
    return count

_afni_suffix_regex = r"((\+(orig|tlrc|acpc))?\.?(nii|HEAD|BRIK)?(.gz|.bz2)?)(\[\d+\])?$"

def prefix(filename):
    ''' strips common fMRI dataset suffixes from filenames '''
    return os.path.split(re.sub(_afni_suffix_regex,"",str(filename)))[1]

def suffix(filename,suffix):
    ''' returns a filenames with ``suffix`` inserted before the dataset suffix '''
    return os.path.split(re.sub(_afni_suffix_regex,"%s\g<1>" % suffix,str(filename)))[1]

def afni_copy(filename):
    ''' creates a ``+orig`` copy of the given dataset and returns the filename as a string '''
    afni_filename = "%s+orig" % prefix(filename)
    if not os.path.exists(afni_filename + ".HEAD"):
        subprocess.call(['3dcalc','-a',str(filename),'-expr','a','-prefix',prefix(filename)])
    return afni_filename

def nifti_copy(filename,copy_prefix=None):
    ''' creates a ``.nii.gz`` copy of the given dataset and returns the filename as a string 
    
    If no prefix is given, will use prefix from ``filename``'''
    nifti_filename = prefix(filename) + ".nii.gz"
    if prefix:
        nifti_filename = prefix(copy_prefix) + ".nii.gz"
    if not os.path.exists(nifti_filename):
        subprocess.call(['3dAFNItoNIFTI',str(filename),'-prefix',nifti_filename])
    return nifti_filename

def is_nifti(filename):
    info = dset_info(filename)
    if 'filetype' in dir(info):
        if info.filetype=='NIFTI':
            return True
        else:
            return False
    else:
        if filename.find('.nii')>0:
            return True
        else:
            return False

def is_dset(filename):
    '''just checks if the filename has the format of an fMRI dataset'''
    afni_dset = '^.*\+(orig|acpc|tlrc)\.?(HEAD|BRIK)?(.gz|.bz)?$'
    nifti_dset = '^.*\.nii(.gz|.bz)?$'
    return (re.match(afni_dset,filename) or re.match(nifti_dset,filename))!=None

class temp_afni_copy:
    ''' used within a ``with`` block, will create a temporary ``+orig`` copy of dataset
    
    When invoked in a ``with`` block, will make a ``+orig`` copy of the given dataset
    and return the filename as a string.
    
    On exiting the block, it will delete the ``+orig`` copy. If out_dsets is a string
    or list of strings, it will try to make nifti copies of all datasets listed 
    and then delete the original version if successful. AFNI datasets should be specified
    with the ``+view``.
    
    Example usage::
        with temp_afni_copy('dataset.nii.gz') as dset_afni:
            do_something_with_a_dset(dset_afni)
        
        with temp_afni_copy('dataset.nii.gz','output_dataset+orig') as dset_afni:
            do_something_with_a_dset(dset_afni,output='output_dataset')
    '''
    def __init__(self,in_dset,out_dsets=None):
        self.in_dset = in_dset
        self.out_dsets = out_dsets

    def __enter__(self):
        self.afni_filenames = []
        if isinstance(self.in_dset,list):
            for dset in self.in_dset:
                self.afni_filenames.append(afni_copy(dset))
            return self.afni_filenames
        else:
            self.afni_filenames.append(afni_copy(self.in_dset))
            return self.afni_filenames[0]

    def __exit__(self, type, value, traceback):
        for afni_filename in self.afni_filenames:   
            for dset in glob.glob(afni_filename + '.*'):
                try:
                    os.remove(dset)
                except OSError:
                    pass
        if self.out_dsets:
            if isinstance(self.out_dsets,basestring):
                self.out_dsets = [self.out_dsets]
            for out_dset in self.out_dsets:
                nifti_dset = nifti_copy(out_dset)
                if os.path.exists(nifti_dset):
                    for dset in [out_dset,out_dset+'.HEAD']:
                        try:
                            os.remove(dset)
                        except OSError:
                            pass

class Decon:
    '''wrapper for AFNI 3dDeconvolve command
    
    Properties:
        :input_dset:        list of input datasets
        :stim_files:        dictionary where keys are used as stimulus labels
                            and the values are taken as 1D files
        :stim_times:        same as stim_files, but used as a stim_times file
        :stim_times_model:  default model to use for each stim_times
        :stim_base:         list of names of stimuli (defined either in stim_files or 
                            stim_times) that should be considered in the baseline instead
                            of full model
        :stim_am1:          list of names of stimuli defined in stim_times that should
                            use the ``-stim_times_AM1`` model
        :stim_am2:          list of names of stimuli defined in stim_times that should
                            use the ``-stim_times_AM2`` model
        :glts:              dictionary where keys are GLT labels, and the value
                            is a symbolic statement
        :mask:              either a mask file, or "auto", which will use "-automask"
        
        Options that are obvious:
            nfirst (default: 3), censor_file, polort (default: 'A'), tout, vout, rout, prefix
            
        opts        = list of extra things to put in the command
        
        **Example:**::
        
            decon = neural.afni.Decon()
            decon.input_dsets = ['dset_run1.nii.gz', 'dset_run2.nii.gz']
            decon.censor_file = 'subject_censor.1D'
            decon.stim_files = {
                'motion_1': 'motion_file.1D[0]',
                'motion_2': 'motion_file.1D[1]',
                'motion_3': 'motion_file.1D[2]',
                'motion_4': 'motion_file.1D[3]',
                'motion_5': 'motion_file.1D[4]',
                'motion_6': 'motion_file.1D[5]'
            }
            decon.stim_base = ['motion_%d' % i for i in range(1,7)]
            decon.stim_times = {
                'stim_a': 'stim_a.stimtimes',
                'stim_b': 'stim_b.stimtimes',
                'stim_c': 'stim_c.stimtimes',
            }
            decon.glts ={
                'a-b': '1*stim_a + -1*stim_b',
                'ab-c': '0.5*stim_a + 0.5*stim_b + -1*stim_c'
            }
            decon.prefix = 'subject_decon.nii.gz'
            decon.run()
    '''
    def __init__(self):
        self.input_dsets=[]
        self.stim_files={}
        self.stim_times={}
        self.stim_times_model = 'GAM'
        self.stim_base = []
        self.stim_am1 = []
        self.stim_am2 = []
        self.censor_file=None
        self.glts={}
        self.opts=[]
        self.nfirst = 3
        self.mask = 'auto'
        self.polort = 'A'
        self.prefix = None
        self.tout = True
        self.vout = True
        self.rout = True
        self.reps = None
        self.TR = None
        self.stim_sds = None
    
    def command_list(self):
        '''returns the 3dDeconvolve command as a list
    
        The list returned can be run by passing it into a subprocess-like command
        (e.g., neural.utils.run())
        '''
        cmd = ['3dDeconvolve']

        cmd += ['-jobs',multiprocessing.cpu_count()]
        cmd += self.opts
        if(len(self.input_dsets)):
            cmd += ['-input'] + self.input_dsets
        else:
            cmd += ['-nodata']
            if self.reps:
                cmd += [str(self.reps)]
                if self.TR:
                    cmd += [str(self.TR)]
        if self.censor_file:
            cmd += ['-censor', self.censor_file]
        cmd += ['-nfirst',str(self.nfirst)]
        if self.mask:
            if self.mask=='auto':
                cmd += ['-automask']
            else:
                cmd += ['-mask',self.mask]
        cmd += ['-polort',str(self.polort)]
        
        cmd += ['-num_stimts',len(self.stim_files)+len(self.stim_times)]
        
        stim_num = 1
        for stim in self.stim_files:
            cmd += ['-stim_file',stim_num,self.stim_files[stim],'-stim_label',stim_num,stim]
            if stim in self.stim_base:
                cmd += ['-stim_base',stim_num]
            stim_num += 1
        
        for stim in self.stim_times:
            opt = '-stim_times'
            if stim in self.stim_am1:
                opt = '-stim_times_AM1'
            if stim in self.stim_am2:
                opt = '-stim_times_AM2'
            cmd += [opt,stim_num,self.stim_times[stim],self.stim_times_model,'-stim_label',stim_num,stim]
            if stim in self.stim_base:
                cmd += ['-stim_base',stim_num]
            stim_num += 1
        
        cmd += ['-num_glt',len(self.glts)]
        
        glt_num = 1
        for glt in self.glts:
            cmd += ['-gltsym','SYM: %s' % self.glts[glt],'-glt_label',glt_num,glt]
            glt_num += 1
        
        if self.tout:
            cmd += ['-tout']
        if self.vout:
            cmd += ['-vout']
        if self.rout:
            cmd += ['-rout']
        
        if self.prefix:
            cmd += ['-bucket', self.prefix]
        
        return [str(x) for x in cmd]
    
    def command_string(self):
        '''returns the 3dDeconvolve command as as string
        
        This command can then be run in something like a shell script
        '''
        return ' '.join(self.command_list())
    
    def run(self,working_directory='.'):
        '''runs 3dDeconvolve through the neural.utils.run shortcut'''
        out = neural.run(self.command_list(),working_directory=working_directory,products=self.prefix)
        if out and out.output:
            stim_sds_list = [x.split() for x in out.output.strip().split('\n\n')]
            self.stim_sds = {}
            for stim in stim_sds_list:
                self.stim_sds[stim[1]] = float(stim[-1])

def tshift(dset,suffix='_tshft',initial_ignore=3):
    neural.run(['3dTshift','-prefix',neural.suffix(dset,suffix),'-ignore',initial_ignore,dset],products=neural.suffix(dset,suffix))

def volreg(dset,suffix='_volreg',base_subbrick=3,tshift=True):
    ''' simple interface to 3dvolreg (recommend looking at align_epi_anat instead of using this) '''
    cmd = ['3dvolreg','-prefix',neural.suffix(dset,suffix),'-base',base_subbrick]
    if tshift:
        cmd += ['-tshift',base_subbrick]
    cmd += [dset]
    neural.run(cmd,products=neural.suffix(dset,suffix))

def affine_align(dset_from,dset_to,skull_strip=True,mask=None,affine_suffix='_aff'):
    ''' interface to 3dAllineate to align anatomies and EPIs '''
    
    dset_ss = lambda dset: os.path.split(suffix(dset,'_ns'))[1]
    def dset_source(dset):      
        if skull_strip==True or skull_strip==dset:
            return dset_ss(dset)
        else:
            return dset
    
    dset_affine = os.path.split(suffix(dset_from,affine_suffix))[1]
    dset_affine_1D = prefix(dset_affine) + '.1D'
        
    if os.path.exists(dset_affine):
        # final product already exists
        return
    
    for dset in [dset_from,dset_to]:
        if skull_strip==True or skull_strip==dset:
            neural.fsl.skull_strip(dset,'_ns')
        
    mask_use = mask
    if mask:
        # the mask was probably made in the space of the original dset_to anatomy,
        # which has now been cropped from the skull stripping. So the lesion mask
        # needs to be resampled to match the corresponding mask
        if skull_strip==True or skull_strip==dset_to:
            neural.run(['3dresample','-master',dset_u(dset_ss(dset)),'-inset',mask,'-prefix',neural.afni.suffix(mask,'_resam')],products=neural.afni.suffix(mask,'_resam'))
            mask_use = neural.afni.suffix(mask,'_resam')
    
    all_cmd = [
        '3dAllineate',
        '-prefix', dset_affine,
        '-base', dset_source(dset_to),
        '-source', dset_source(dset_from),
        '-cost', 'lpa',
        '-1Dmatrix_save', dset_affine_1D,
        '-autoweight',
        '-cmass'
    ]
    
    if mask:
        all_cmd += ['-emask', mask_use]
    
    neural.run(all_cmd,products=dset_affine)
    

def qwarp_align(dset_from,dset_to,skull_strip=True,mask=None,affine_suffix='_aff',qwarp_suffix='_qwarp'):
    '''aligns ``dset_from`` to ``dset_to`` using 3dQwarp
    
    Will run ``3dSkullStrip`` (unless ``skull_strip`` is ``False``), ``3dUnifize``,
    ``3dAllineate``, and then ``3dQwarp``. This method will add suffixes to the input
    dataset for the intermediate files (e.g., ``_ss``, ``_u``). If those files already
    exist, it will assume they were intelligently named, and use them as is
    
    :skull_strip:       If True/False, turns skull-stripping of both datasets on/off.
                        If a string matching ``dset_from`` or ``dset_to``, will only
                        skull-strip the given dataset
    :mask:              Applies the given mask to the alignment. Because of the nature
                        of the alignment algorithms, the mask is **always** applied to
                        the ``dset_to``. If this isn't what you want, you need to reverse
                        the transform and re-apply it (e.g., using :meth:`qwarp_invert` 
                        and :meth:`qwarp_apply`). If the ``dset_to`` dataset is skull-stripped,
                        the mask will also be resampled to match the ``dset_to`` grid.
    :affine_suffix:     Suffix applied to ``dset_from`` to name the new dataset, as well as
                        the ``.1D`` file.
    :qwarp_suffix:      Suffix applied to the final ``dset_from`` dataset. An additional file
                        with the additional suffix ``_WARP`` will be created containing the parameters
                        (e.g., with the default ``_qwarp`` suffix, the parameters will be in a file with
                        the suffix ``_qwarp_WARP``)
    
    The output affine dataset and 1D, as well as the output of qwarp are named by adding
    the given suffixes (``affine_suffix`` and ``qwarp_suffix``) to the ``dset_from`` file
    
    If ``skull_strip`` is a string instead of ``True``/``False``, it will only skull strip the given
    dataset instead of both of them
    
    # TODO: currently does not work with +tlrc datasets because the filenames get mangled
    '''
    
    dset_ss = lambda dset: os.path.split(suffix(dset,'_ns'))[1]
    dset_u = lambda dset: os.path.split(suffix(dset,'_u'))[1]
    def dset_source(dset):      
        if skull_strip==True or skull_strip==dset:
            return dset_ss(dset)
        else:
            return dset
    
    dset_affine = os.path.split(suffix(dset_from,affine_suffix))[1]
    dset_affine_1D = prefix(dset_affine) + '.1D'
    dset_qwarp = os.path.split(suffix(dset_from,qwarp_suffix))[1]
    
    if os.path.exists(dset_qwarp):
        # final product already exists
        return
    
    affine_align(dset_from,dset_to,skull_strip,mask,affine_suffix)
      
    for dset in [dset_from,dset_to]:  
        neural.run([
            '3dUnifize',
            '-prefix', dset_u(dset_source(dset)),
            '-input', dset_source(dset)
        ],products=[dset_u(dset_source(dset))])
    
    mask_use = mask
    if mask:
        # the mask was probably made in the space of the original dset_to anatomy,
        # which has now been cropped from the skull stripping. So the lesion mask
        # needs to be resampled to match the corresponding mask
        if skull_strip==True or skull_strip==dset_to:
            neural.run(['3dresample','-master',dset_u(dset_ss(dset)),'-inset',mask,'-prefix',neural.afni.suffix(mask,'_resam')],products=neural.afni.suffix(mask,'_resam'))
            mask_use = neural.afni.suffix(mask,'_resam')
    
    warp_cmd = [
        '3dQwarp',
        '-prefix', dset_qwarp,
        '-duplo', '-useweight', '-blur', '0', '3',
        '-iwarp',
        '-base', dset_u(dset_source(dset_to)),
        '-source', dset_affine
    ]
    
    if mask:
        warp_cmd += ['-emask', mask_use]
    
    neural.run(warp_cmd,products=dset_qwarp)

def qwarp_apply(dset_from,dset_warp,affine=None,warp_suffix='_warp',master='WARP',interp=None):
    '''applies the transform from a previous qwarp
    
    Uses the warp parameters from the dataset listed in 
    ``dset_warp`` (usually the dataset name ends in ``_WARP``) 
    to the dataset ``dset_from``. If a ``.1D`` file is given
    in the ``affine`` parameter, it will be applied simultaneously
    with the qwarp.
    
    If the parameter ``interp`` is given, will use as interpolation method,
    otherwise it will just use the default (currently wsinc5)
    
    Output dataset with have the ``warp_suffix`` suffix added to its name
    '''
    out_dset = os.path.split(suffix(dset_from,warp_suffix))[1]
    dset_from_info = dset_info(dset_from)
    dset_warp_info = dset_info(dset_warp)
    if(dset_from_info.orient!=dset_warp_info.orient):
        # If the datasets are different orientations, the transform won't be applied correctly
        nl.run(['3dresample','-orient',dset_warp_info.orient,'-prefix',suffix(dset_from,'_reorient'),'-inset',dset_from])
        dset_from = suffix(dset_from,'_reorient')
    cmd = [
        '3dNwarpApply',
        '-nwarp', dset_warp]
    if affine:
        cmd += ['-affter', affine]
    cmd += [
        '-source', dset_from,
        '-master',master,
        '-prefix', out_dset
    ]
    
    if interp:
        cmd += ['-interp',interp]
    
    neural.run(cmd,products=out_dset)

def qwarp_invert(warp_param_dset,output_dset,affine_1Dfile=None):
    '''inverts a qwarp (defined in ``warp_param_dset``) (and concatenates affine matrix ``affine_1Dfile`` if given)
    outputs the inverted warp + affine to ``output_dset``'''
    
    cmd = ['3dNwarpCat','-prefix',output_dset]
    
    if affine_1Dfile:
        cmd += ['-warp1','INV(%s)' % affine_1Dfile, '-warp2','INV(%s)' % warp_param_dset]
    else:
        cmd += ['-warp1','INV(%s)' % warp_param_dset]
            
    neural.run(cmd,products=output_dset)

def skull_strip(dset,out_suffix='_ns'):
    ''' runs 3dSkullStrip (I would recommend looking at fsl.bet instead)'''
    neural.run([
        '3dSkullStrip',
        '-input', dset,
        '-prefix', suffix(dset,out_suffix),
        '-niter', '400',
        '-ld', '40'
    ],products=suffix(dset,out_suffix))

def align_epi_anat(anatomy,epi_dsets):
    ''' aligns epis to anatomy using ``align_epi_anat.py`` script
    
    :epi_dsets:       can be either a string or list of strings of the epi child datasets
    
    The default output suffix is "_al"
    '''
    
    if isinstance(epi_dsets,basestring):
        epi_dsets = [epi_dsets]
        
    if len(epi_dsets)==0 or epi_dsets[0]=='' or os.path.exists(suffix(epi_dsets[0],'_al')):
        return
    
    anatomy_use = anatomy
    if is_nifti(anatomy):
        anatomy_use = afni_copy(anatomy)
    epi_dsets_use = []
    for dset in epi_dsets:
        if is_nifti(dset):
            epi_dsets_use.append(afni_copy(dset))
        else:
            epi_dsets_use.append(dset)
    
    cmd = ["align_epi_anat.py", "-epi2anat", "-anat", anatomy_use, "-epi_base", "5", "-epi", epi_dsets_use[0]]
    if len(epi_dsets_use)>1:
        cmd += ['-child_epi'] + epi_dsets_use[1:]
    out = neural.run(cmd)
    
    for dset in ([anatomy] + epi_dsets):
        if is_nifti(dset):
            if dset!=anatomy:
                dset_nifti = nifti_copy(prefix(dset)+'_al+orig')
                nl.run(['gzip',dset_nifti])
            if dset==anatomy or os.path.exists(suffix(dset,'_al')):
                for s in ['_al+orig.HEAD','_al+orig.BRIK*','+orig.HEAD','+orig.BRIK*']:
                    for d in glob.glob(prefix(dset) + s):
                        if os.path.exists(d):
                            os.remove(d)

def auto_polort(dset):
    '''a copy of 3dDeconvolve's ``-polort A`` option'''
    info = dset_info(dset)
    return 1 + round(info.reps/150.0)

class AFNI_Censor_TooManyOutliers (RuntimeError):
    pass

def create_censor_file(input_dset,out_prefix=None,fraction=0.1,clip_to=0.1,max_exclude=0.3):
    '''create a binary censor file using 3dToutcount
    
    :input_dset:        the input dataset
    :prefix:            output 1D file (default: ``prefix(input_dset)`` + ``.1D``)
    :fraction:          censor a timepoint if proportional of outliers in this
                        time point is greater than given value
    :clip_to:           keep the number of time points censored under this proportion
                        of total reps. If more time points would be censored, 
                        it will only pick the top ``clip_to*reps`` points
    :max_exclude:       if more time points than the given proportion of reps are excluded for the 
                        entire run, throw an exception -- something is probably wrong
    '''
    polort = auto_polort(input_dset)
    info = dset_info(input_dset)
    outcount = [float(x.split()[0]) for x in subprocess.check_output(['3dToutcount','-fraction','-automask','-polort',str(polort),str(input_dset)]).split('\n')]
    binary_outcount = [x<fraction for x in outcount]
    perc_outliers = sum(binary_outcount)/float(info.reps)
    if max_exclude and perc_outliers > max_exclude:
        raise AFNI_Censor_TooManyOutliers('Found %f outliers in dset %s' % (perc_outliers,input_dset))
    if clip_to:
        while perc_outliers > clip_to:
            best_outlier = min([(outcount[i],i) for i in range(len(outcount)) if binary_outcount[i]])
            binary_outcount[best_outlier[1]] = False
            perc_outliers = sum(binary_outcount)/float(info.reps)
    if not out_prefix:
        out_prefix = prefix(input_dset) + '.1D'
    with open(out_prefix,'w') as f:
        f.write('\n'.join([str(int(x)) for x in binary_outcount]))



'''methods to implement lightweight job distribution across multiple machines

All of these functions are made on the assumption that all clients' filesystems
are identical, for example on a shared folder'''
import zmq
import json,time
from datetime import datetime
import neural as nl

default_port = 8765
packet_expire = 60.0 # seconds
_key = 'scheduler'
info = None

class Server:
    def __init__(self,name=None,address='localhost',port=default_port,password=None):
        self.name = name
        self.address = address
        self.port = port
        self.password = password
    
    def start_server(self):
        c = zmq.Context()
        sock = c.socket(zmq.REP)
        sock.bind('tcp://*:%s' % self.port)
        
        try:
            while True:
                msg = sock.recv()
                try:
                    obj = json.loads(msg)
                    if 'key' not in obj or obj['key']!=_key:
                        raise ValueError
                    if self.password:
                        if 'password' not in obj or 'time' not in obj:
                            raise ValueError
                        if nl.hash_str(self.password + str(obj['time']))!= obj['password']:
                            raise ValueError
                except ValueError:
                    # not a valid packet
                    sock.send('ERR')
                    continue
                else:
                    if 'task' in obj:
                        if obj['task']=='info':
                            sock.send(json.dumps({'name':self.name,'address':self.address,'port':self.port}))
                            continue
                    
                        if obj['task']=='job':
                            print 'Being asked to run command: ' + str(obj['command'])
                            cmd = [obj['command'],None,'.']
                            if 'products' in obj:
                                cmd[1] = obj['products']
                            if 'working_directory' in obj:
                                cmd[2] = obj['working_directory']
                            output = nl.run(*cmd)
                            sock.send(json.dumps({'output':output.__dict__()}))
                            continue
                
                    sock.send('OK')
        except KeyboardInterrupt:
            c.destroy()

def _send_raw(msg,address='localhost',port=default_port):
    c = zmq.Context()
    sock = c.socket(zmq.REQ)
    sock.connect('tcp://%s:%s' % (address,port))
    sock.send(msg)
    return sock.recv()  

def server_info(address='localhost',port=default_port):
    info_req = json.dumps({'key':_key,'task':'info'})
    rep = _send_raw(info_req,address,port)
    info = ServerInfo()
    try:
        rep_dict = json.loads(rep)
        for key in rep_dict:
            setattr(info,key,rep_dict[key])
    except (ValueError,KeyError):
        return None
    return info

class Job:
    def __init__(self,command=[],products=None,working_directory='.'):
        self.command = command
        self.working_directory = working_directory
        self.products = products
        self.result = None

def send_job(job,address='localhost',port=default_port,password=None):
    job_req_dict = dict({'key':_key,'task':'job','time':time.time()}.items() + job.__dict__.items())
    if password:
        job_req_dict['password'] = nl.hash_str(password + str(job_req_dict['time']))
    job_req = json.dumps(job_req_dict)
    rep = _send_raw(job_req,address,port)
    try:
        rep_dict = json.loads(rep)
        for key in rep_dict:
            if key=='output':
                output = nl.utils.RunResult()
                for okey in output:
                    setattr(output,okey,output[okey])
                job.output = output
            else:
                setattr(job,key,rep_dict[key])
    except (ValueError,KeyError):
        return None
    return job

class Scheduler:
    def __init__(self):
        self.servers = [{'address':'local'}]
    
    def add_server(self,address,port=default_port,password=None,speed=None,valid_times=None,invalid_times=None):
        '''
        :address:           remote address of server, or special string ``local`` to
                            run the command locally
        :valid_times:       times when this server is available, given as a list
                            of tuples of 2 strings of form "HH:MM" that define the
                            start and end times. Alternatively, a list of 7 lists can
                            be given to define times on a per-day-of-week basis
        E.g.,::
        
            [('4:30','14:30'),('17:00','23:00')]
            # or
            [
                [('4:30','14:30'),('17:00','23:00')],       # S
                [('4:30','14:30'),('17:00','23:00')],       # M
                [('4:30','14:30'),('17:00','23:00')],       # T
                [('4:30','14:30'),('17:00','23:00')],       # W
                [('4:30','14:30'),('17:00','23:00')],       # R
                [('4:30','14:30'),('17:00','23:00')],       # F
                [('4:30','14:30'),('17:00','23:00')]        # S
            ]
        
        :invalid_times:     uses the same format as ``valid_times`` but defines times
                            when the server should not be used
        '''
        for t in [valid_times,invalid_times]:
            if t:
                if not (self._is_list_of_tuples(t) or self._is_list_of_tuples(t,True)):
                    raise ValueError('valid_times and invalid_times must either be lists of strings or lists')
        self.servers.append({
            'address':address,
            'port':port,
            'password':password,
            'speed':speed,
            'valid_times':valid_times,
            'invalid_times':invalid_times
        })
    
    def _is_list_of_tuples(self,l,is_list=False):
        if not isinstance(l,list):
            return False
        for x in l:
            if is_list:
                if not isinstance(x,list):
                    return False
                for y in x:
                    if not isinstance(y,tuple):
                        return False
                    for z in y:
                        if not isinstance(z,basestring):
                            return False
            else:
                if not isinstance(x,tuple):
                    return False
                for z in x:
                    if not isinstance(z,basestring):
                        return False
        return True
    
    def _time_is_inbetween(self,t,tt):
        tt_from = [int(x) for x in tt[0].split(':')]
        tt_to = [int(x) for x in tt[1].split(':')]
        if ((t.hour>tt_from[0] or (t.hour==tt_from[0] and t.minute>tt_from[1])) and
            (t.hour<tt_to[0] or (t.hour==tt_to[0] and t.minute<tt_to[1]))):
            return True
        return False
    
    def choose_server(self):
        valid_servers = []
        now = datetime.now()
        for server in self.servers:
            valid = True
            if 'valid_times' in server and server['valid_times']:
                times = server['valid_times']
                if self._is_list_of_tuples(times,is_list=True):
                    wday = (now.weekday() + 1) % 7  # Make Sunday==0
                    times = server['valid_times'][wday]
                valid = False
                for t in times:
                    if self._time_is_inbetween(now,t):
                        valid = True
                        break
            if 'invalid_times' in server and server['invalid_times']:
                times = server['invalid_times']
                if self._is_list_of_tuples(times,is_list=True):
                    wday = (now.weekday() + 1) % 7  # Make Sunday==0
                    times = server['valid_times'][wday]
                for t in times:
                    if self._time_is_inbetween(now,t):
                        valid = False
                        break
            if valid:
                valid_servers.append(server)
        sorted_servers = sorted(valid_servers,key=lambda x: x['speed'] if 'speed' in x and x['speed'] else 0)
        return sorted_servers[-1]

scheduler = Scheduler()

def _new_run(command,products=None,working_directory='.',force_local=False):
    server = scheduler.choose_server()
    if force_local or server['address']=='local':
        return nl.utils.run(command,products,working_directory,force_local)
    else:
        job = Job(command,products,working_directory)
        if 'password' in server:
            job_return = send_job(job,server['address'],server['port'],server['password'])
        else:
            job_return = send_job(job,server['address'],server['port'])
        return job_return.output

nl.run = _new_run


''' Provides a centralized notification function

The functions will try to detect the specific system and situation to tailor
the notification to the most appropriate method
'''
import sys,os,io
import platform
import random,string
import smtplib
from email.mime.text import MIMEText
import neural.term

class SMTPServer:
    def __init__(self,from_addr,server,username=None,password=None,SSL=False):
        self.from_addr = from_addr
        self.server = server
        self.username = username
        self.password = password
        self.SSL = SSL

default_SMTP = None

def enable_email(from_email,server,username=None,password=None,SSL=False):
    global default_SMTP
    default_SMTP = SMTPServer(from_email,server,username,password,SSL)

def email(to,msg,subject='Neural notification'):
    if default_SMTP==None:
        raise RuntimeError('Email not enabled, must run ``enable_email`` first!')
    msg = MIMEText(msg)
    msg['Subject'] = subject
    msg['From'] = default_SMTP.from_addr
    msg['To'] = to
    if default_SMTP.SSL:
        s = smtplib.SMTP_SSL(default_SMTP.server)
    else:
        s = smtplib.SMTP(default_SMTP.server)
    if default_SMTP.username:
        s.login(default_SMTP.username,default_SMTP.password)
    s.sendmail(to,default_SMTP.from_addr,msg.as_string())
    s.quit()

# Log levels:
class level:
    debug, informational, warning, error, critical = range(5)

interactive_enabled = False
email_digest = None

#! variable to hold list of nested notifications
_notify_tree = []

class notify:
    def __init__(self,text,log=False,email=False,level=level.informational):
        '''notify user or log information
    
        :text:      the content of the message
        :log:       if ``True``, will save the message
        :email:     add to email digests
        :level:     priority of message, should be '''
        self.text = text
        self.log = log
        self.email = email
        self.level = level
        self.id = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(6))
        if interactive_enabled:
            notify_interactive(self)
        else:
            notify_normal(self)
        return
    
    def __enter__(self):
        _notify_tree.append(self)
    
    def __exit__(self, type, value, traceback):
        _notify_tree.pop()

def notify_interactive(n):
    pass

def notify_normal(n):
    prefix = ''
    if(len(_notify_tree) > 0):
        prefix += '  '*len(_notify_tree) + '- '
    n_prefixed = prefix + n.text + '\n'
    try:
        if os.isatty(sys.stderr.fileno()):
            color = 'green'
            if n.level >= level.warning:
                color = 'yellow'
            if n.level >= level.error:
                color = 'red'
            n_prefixed = neural.term.color(n_prefixed,color)
    except io.UnsupportedOperation:
        pass
    sys.stderr.write(n_prefixed)
    sys.stderr.flush()

### Email notification

class email_digest:
    def __init__(self,email_address,level=None):
        '''start an email digest
        
        Will try to capture all notifications sent within the following code block that are
        sent with ``email=True``. In addition, if ``level`` is given, will only collect notifications
        with the given level or higher'''
        pass
    
    def __enter__(self):
        pass
    
    def __exit__(self, type, value, traceback):
        pass
    

### Platform-specific methods:

if platform.system() == 'Darwin':
    if int(platform.release().split(".")[0])>=12:
        try:
            from pync import Notifier
        except ImportError: 
            pass
        else:
            def notify_mountainlion(n):
                notify_normal(n)
                return
                group_id = '%d-%s' % (os.getpid(),n.id)
                text = n.text
                if len(_notify_tree):
                    temp_tree = [x.text for x in _notify_tree] + [text]
                    text = '\n'.join(['%s%s' % ('  '*i,temp_tree[i]) for i in xrange(len(temp_tree))])
                    group_id = '%d-%s' % (os.getpid(),_notify_tree[0].id)
                    Notifier.remove(group_id)
                Notifier.notify(text,title='Jarvis',group=group_id)
            notify_interactive = notify_mountainlion
            interactive_enabled = True

try:
    in_ipython = isinstance(sys.stdout,IPython.kernel.zmq.iostream.OutStream)
except NameError:
    pass
else:
    if in_ipython:
        # We're inside iPython Notebook
        try:
            import IPython.display
        except ImportError:
            pass
        else:
            def notify_ipython_html(text):
                html = IPython.display.HTML( '<b>:: %s</b>' % text)
                IPython.display.display_html(html)
            notify_interactive = notify_ipython_html
            interactive_enabled = True


''' This module contains generic helper functions

Many of these are imported into the root-level of the package to make accessing them easier '''
import neural
#from notify import notify
import os,subprocess
import datetime
import hashlib
import zlib, base64

#! A list of archives this library understands
archive_formats = {
    'zip': {'suffix':'zip', 'command':lambda output,filename: ['unzip','-o','-d',output,filename]},
    'tarball': {'suffix':('tar.gz','tgz'), 'command':lambda output,filename: ['tar','zx','-C',output,'-f',filename]},
    'tarball-bzip': {'suffix':('tar.bz2','tbz'), 'command':lambda output,filename: ['tar','jx','-C',output,'-f',filename]},
    '7zip': {'suffix':'7z', 'command':lambda output,filename: ['7z','x','-y','-o%s' % output, filename]}
}

def is_archive(filename):
    '''returns boolean of whether this filename looks like an archive'''
    for archive in archive_formats:
        if filename.endswith(archive_formats[archive]['suffix']):
            return True
    return False

def archive_basename(filename):
    '''returns the basename (name without extension) of a recognized archive file'''
    for archive in archive_formats:
        if filename.endswith(archive_formats[archive]['suffix']):
            return filename.rstrip('.' + archive_formats[archive]['suffix'])
    return False

def unarchive(filename,output_dir='.'):
    '''unpacks the given archive into ``output_dir``'''
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    for archive in archive_formats:
        if filename.endswith(archive_formats[archive]['suffix']):
            return subprocess.call(archive_formats[archive]['command'](output_dir,filename))==0
    return False

def flatten(nested_list):
    '''converts a list-of-lists to a single flat list'''
    return_list = []
    for i in nested_list:
        if isinstance(i,list):
            return_list += flatten(i)
        else:
            return_list.append(i)
    return return_list

class run_in:
    '''temporarily changes into another directory
    
    If the directory name you pass doesn't exist, but matches the current directory
    you are in, it will silently ignore your silliness. Otherwise, it will raise an
    ``OSError``. If the argument ``create`` is True, it will create the directory instead
    of raising an error.
    
    Example::   
        with run_in('another_directory'):
            do_some_stuff_there()
    '''
    def __init__(self,working_directory,create=False):
        self.working_directory = working_directory
        self.create = create

    def __enter__(self):
        self.old_cwd = os.getcwd()
        if self.working_directory:
            if os.path.exists(self.working_directory):
                os.chdir(self.working_directory)
            elif self.old_cwd.endswith(self.working_directory.rstrip('/')):
                # Silly, we're already in that directory
                pass
            else:
                if self.create:
                    os.makedirs(self.working_directory)
                    os.chdir(self.working_directory)
                else:
                    raise IOError('Attempting to run_in the non-existent directory "%s"' % self.working_directory)
    
    def __exit__(self, type, value, traceback):
        os.chdir(self.old_cwd)

class RunResult:
    '''result of calling :meth:`run`
    
    when used as a string, will try to reasonably return the filename of the primary output
    of the command (if known)'''    
    def __init__(self,output=None,return_code=None,output_filename=None):
        self.output_filename = output_filename
        self.output = output
        self.return_code = return_code
    
    def __str__(self):
        return self.output_filename

def run(command,products=None,working_directory='.',force_local=False):
    '''wrapper to run external programs
    
    :command:           list containing command and parameters 
                        (formatted the same as subprocess; must contain only strings)
    :products:          string or list of files that are the products of this command
                        if all products exist, the command will not be run, and False returned
    :working_directory: will chdir to this directory
    :force_local:       when used with :module:`neural.scheduler`, setting to ``True`` will disable
                        all job distribution functions
    
    Returns result in form of :class:`RunResult`
    '''
    with run_in(working_directory):
        if products:
            if isinstance(products,basestring):
                products = [products]
            if all([os.path.exists(x) for x in products]):
                return False
        
        command = flatten(command)
        command = [str(x) for x in command]
        with neural.notify('Running %s...' % command[0]):
            out = None
            returncode = 0
            try:
                out = subprocess.check_output(command)
            except subprocess.CalledProcessError, e:
                neural.notify('''ERROR: %s returned a non-zero status

----COMMAND------------
%s
-----------------------

                    
----OUTPUT-------------
%s
-----------------------
Return code: %d
''' % (command[0],' '.join(command),e.output,e.returncode),level=neural.level.error)
                returncode = e.returncode
            result = RunResult(out,returncode)
            if products and returncode==0:
                result.output_filename = products[0]
            return result

def log(fname,msg):
    ''' generic logging function '''
    with open(fname,'a') as f:
        f.write(datetime.datetime.now().strftime('%m-%d-%Y %H:%M:\n') + msg + '\n')

def hash(filename):
    '''returns string of MD5 hash of given filename'''
    buffer_size = 10*1024*1024
    m = hashlib.md5()
    with open(filename) as f:
        buff = f.read(buffer_size)
        while len(buff)>0:
            m.update(buff)
            buff = f.read(buffer_size)          
    dig = m.digest()
    return ''.join(['%x' % ord(x) for x in dig])

def hash_str(string):
    '''returns string of MD5 hash of given string'''
    m = hashlib.md5()
    m.update(string)
    dig = m.digest()
    return ''.join(['%x' % ord(x) for x in dig])

class simple_timer:
    '''a simple way to time a single run of a function
    
    Example::
        with simple_timer():
            do_stuff()
    '''
    def __init__(self):
        self.start_time = None
    
    def __enter__(self):
        self.start_time = datetime.datetime.now()
        print 'timer start time: %s' % self.start_time.strftime('%Y-%m-%d %H:%M:%S')
    
    def __exit__(self, type, value, traceback):
        self.end_time = datetime.datetime.now()
        print 'timer end time: %s' % self.end_time.strftime('%Y-%m-%d %H:%M:%S')
        print 'time elapsed: %s' % str(self.end_time-self.start_time)

def factor(n):
    '''return set of all prime factors for a number'''
    return set(reduce(list.__add__, ([i, n//i] for i in range(1, int(n**0.5) + 1) if n % i == 0)))

def compress(data, level=9):
    ''' return compressed, ASCII-encoded string '''
    return base64.encodestring(zlib.compress(data,9))

def decompress(data):
    ''' return uncompressed string '''
    return zlib.decompress(base64.decodestring(data))



'''methods to analyze DICOM format images'''

from __future__ import absolute_import
import subprocess,re,os,multiprocessing,glob
import neural as nl
import string
# No, I'm not importing myself... this is actually the pydicom library
import dicom as pydicom
from fuzzywuzzy import process

def is_dicom(filename):
    '''returns Boolean of whether the given file has the DICOM magic number'''
    try:
        with open(filename) as f:
            d = f.read(132)
            return d[128:132]=="DICM"
    except:
        return False

_dicom_hdr = 'dicom_hdr'

class DicomInfo:
    '''container for header information from DICOM file
    
    Header frames are dictionaries with the following values:
    
    :addr:      tuple of (group,element) tags in hex
    :label:     description
    :offset:    offset of data
    :size:      size of data
    :value:     data value
    '''
    
    def __init__(self,frames=[],sex_info={},slice_times=[]):
        self.raw_frames = frames    #! list of dictionaries containing information on each frame
        self.sex_info = sex_info    #! dictinary with Siemen's extra info fields
        self.slice_times = slice_times #! list of Siemen's slice timing information     
    
    def addr(self,address):
        '''returns dictionary with frame information for given address (a tuple of two hex numbers)'''
        if isinstance(address,basestring):
            # If you just gave me a single string, assume its "XXXX XXXX"
            addr = address.split()
        else:
            addr = list(address)            
        # Convert to actual hex if you give me strings
        for i in xrange(len(addr)):
            if isinstance(addr[i],basestring):
                addr[i] = int(addr[i],16)
        for frame in self.raw_frames:
            if frame['addr']==address:
                return frame
    
    def label(self,label):
        '''returns dictionary with frame information for a given text label'''
        for frame in self.raw_frames:
            if frame['label']==label:
                return frame

def info(filename):
    '''returns a DicomInfo object containing the header information in ``filename``'''
    try:
        out = subprocess.check_output([_dicom_hdr,'-sexinfo',filename])
    except subprocess.CalledProcessError:
        return None
    slice_timing_out = subprocess.check_output([_dicom_hdr,'-slice_times',filename])
    slice_timing = [float(x) for x in slice_timing_out.strip().split()[5:]]
    frames = []
    for frame in re.findall(r'^(\w{4}) (\w{4})\s+(\d+) \[(\d+)\s+\] \/\/(.*?)\/\/(.*?)$',out,re.M):
        new_frame = {}
        new_frame['addr'] = (int(frame[0],16),int(frame[1],16))
        new_frame['size'] = int(frame[2])
        new_frame['offset'] = int(frame[3])
        new_frame['label'] = frame[4].strip()
        new_frame['value'] = frame[5].strip()
        frames.append(new_frame)
    sex_info = {}
    for i in re.findall(r'^(.*?)\s+= (.*)$',out,re.M):
        sex_info[i[0]] = i[1]

    return DicomInfo(frames,sex_info,slice_timing)

def info_for_tags(filename,tags):
    '''return a dictionary for the given ``tags`` in the header of the DICOM file ``filename``
    
    ``tags`` is expected to be a list of tuples that contains the DICOM address in hex values.
    
    basically a rewrite of :meth:`info` because it's so slow. This is a lot faster and more reliable'''
    d = pydicom.read_file(filename)
    return {k:d[k].value for k in tags if k in d}

def scan_dir(dirname,tags=None,md5_hash=False):
    '''scans a directory tree and returns a dictionary with files and key DICOM tags
    
    return value is a dictionary absolute filenames as keys and with dictionaries of tags/values
    as values
    
    the param ``tags`` is the list of DICOM tags (given as tuples of hex numbers) that 
    will be obtained for each file. If not given,
    the default list is:
    
    :0008 0021:     Series date
    :0008 0031:     Series time
    :0008 103E:     Series description
    :0008 0080:     Institution name
    :0010 0020:     Patient ID
    :0028 0010:     Image rows
    :0028 0011:     Image columns
    
    If the param ``md5_hash`` is ``True``, this will also return the MD5 hash of the file. This is useful
    for detecting duplicate files
    '''
    if tags==None:
        tags = [
            (0x0008, 0x0021),
            (0x0008, 0x0031),
            (0x0008, 0x103E),
            (0x0008, 0x0080),
            (0x0010, 0x0020),
            (0x0028, 0x0010),
            (0x0028, 0x0011),
        ]
    
    return_dict = {}    
    
    for root,dirs,files in os.walk(dirname):
        for filename in files:
            fullname = os.path.join(root,filename)
            if is_dicom(fullname):
                return_dict[fullname] = info_for_tags(fullname,tags)
                if md5_hash:
                    return_dict[fullname]['md5'] = nl.hash(fullname)
    return return_dict

valid = '_.' + string.ascii_letters + string.digits
def scrub_fname(fname):
    return ''.join(c for c in fname.replace(' ','_') if c in valid).replace('__','_')
    
def find_dups(file_dict):
    '''takes output from :meth:`scan_dir` and returns list of duplicate files'''
    found_hashes = {}
    for f in file_dict:
        if file_dict[f]['md5'] not in found_hashes:
            found_hashes[file_dict[f]['md5']] = []
        found_hashes[file_dict[f]['md5']].append(f)
    final_hashes = dict(found_hashes)
    for h in found_hashes:
        if len(found_hashes[h])<2:
            del(final_hashes[h])
    return final_hashes.values()

def cluster_files(file_dict):
    '''takes output from :meth:`scan_dir` and organizes into lists of files with the same tags
    
    returns a dictionary where values are a tuple of the unique tag combination and values contain
    another dictionary with the keys ``info`` containing the original tag dict and ``files`` containing
    a list of files that match'''
    return_dict = {}
    for filename in file_dict:
        info_dict = dict(file_dict[filename])
        if 'md5' in info_dict:
            del(info_dict['md5'])
        dict_key = tuple(sorted([file_dict[filename][x] for x in info_dict]))
        if dict_key not in return_dict:
            return_dict[dict_key] = {'info':info_dict,'files':[]}
        return_dict[dict_key]['files'].append(filename)
    return return_dict

def max_diff(dset_a,dset_b):
    '''calculates maximal voxel-wise difference in datasets
    
    Useful for checking if datasets have the same data. For example, if the maximum difference is
    < 1.0, they're probably the same dataset'''
    for dset in [dset_a,dset_b]:
        if not os.path.exists(dset):
            raise IOError('Could not find file: %s' % dset)
    try:
        with open(os.devnull,'w') as null:
            return float(subprocess.check_output(['3dBrickStat','-max','3dcalc( -a %s -b %s -expr abs(a-b) )' %(dset_a,dset_b)],stderr=null).split()[0])
    except subprocess.CalledProcessError:
        return float('inf')

def _create_dset_dicom(directory):
    with nl.notify('Trying to create datasets from %s' % directory):
        d = str(directory)
        while d.endswith('/'):
            d = d[:1]
    
        if not os.path.exists(directory):
            nl.notify('Error: could not find %s',level=nl.level.error)
            return False
        
        out_file = '%s.nii' % nl.afni.prefix(d)
        return_val = out_file + '.gz'
        try:
            out = subprocess.check_output([
            'Dimon',
            '-infile_prefix','%s/' % directory,
            '-dicom_org', '-GERT_Reco', 
            '-gert_to3d_prefix', nl.afni.prefix(d),
            '-gert_create_dataset', '-gert_write_as_nifti', '-gert_quit_on_err',
            '-max_images','100000',
            '-quit'],stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError, e:
            nl.notify('Error: Dimon returned an error while creating dataset',level=nl.level.error)
            return_val = False
        
        if os.path.exists(out_file):
            nl.run(['gzip',out_file])
        else:
            return_val = False
        
        for junk in ['dimon.files.run.*','GERT_Reco_dicom_*']:
            for fname in glob.glob(junk):
                try:
                    os.remove(fname)
                except IOError:
                    pass
        return return_val

def create_dset(directory):
    '''tries to autocreate a dataset from images in the given directory'''
    return _create_dset_dicom(directory)
    # Add more options for GE I-files, and other non-DICOM data formats

def organize_dir(orig_dir):
    '''scans through the given directory and organizes DICOMs that look similar into subdirectories
    
    output directory is the ``orig_dir`` with ``-sorted`` appended to the end'''
    
    tags = [
        (0x10,0x20),    # Subj ID
        (0x8,0x21),     # Date
        (0x8,0x31),     # Time
        (0x8,0x103e)    # Descr
    ]
    files = scan_dir(orig_dir,tags=tags,md5_hash=True)
    dups = find_dups(files)
    for dup in dups:
        nl.notify('Found duplicates of %s...' % dup[0])
        for each_dup in dup[1:]:
            nl.notify('\tdeleting %s' % each_dup)
            try:
                os.remove(each_dup)
            except IOError:
                nl.notify('\t[failed]')
            del(files[each_dup])
    
    clustered = cluster_files(files)
    output_dir = '%s-sorted' % orig_dir
    for key in clustered:
        if (0x8,0x31) in clustered[key]['info']:
            clustered[key]['info'][(0x8,0x31)] = str(int(float(clustered[key]['info'][(0x8,0x31)])))
        for t in tags:
            if t not in clustered[key]['info']:
                clustered[key]['info'][t] = '_'
        run_name = '-'.join([scrub_fname(str(clustered[key]['info'][x])) for x in tags])+'-%d_images' %len(clustered[key]['files'])
        run_dir = os.path.join(output_dir,run_name)
        nl.notify('Moving files into %s' % run_dir)
        try:
            if not os.path.exists(run_dir):
                os.makedirs(run_dir)
        except IOError:
            nl.notify('Error: failed to create directory %s' % run_dir)
        else:
            for f in clustered[key]['files']:
                try:
                    dset_fname = os.path.split(f)[1]
                    if dset_fname[0]=='.':
                        dset_fname = '_' + dset_fname[1:]
                    os.rename(f,os.path.join(run_dir,dset_fname))
                except (IOError, OSError):
                    pass
    for r,ds,fs in os.walk(output_dir,topdown=False):
        for d in ds:
            dname = os.path.join(r,d)
            if len(os.listdir(dname))==0:
                os.remove(dname)
                
def classify(label_dict,image_fname=None,image_label=None):
    '''tries to classify a DICOM image based on known string patterns (with fuzzy matching)
    
    Takes the label from the DICOM header and compares to the entries in ``label_dict``. If it finds something close
    it will return the image type, otherwise it will return ``None``. Alternatively, you can supply your own string, ``image_label``,
    and it will try to match that.
    
    ``label_dict`` is a dictionary where the keys are dataset types and the values are lists of strings that match that type.
    For example::
    
        {
            'anatomy': ['SPGR','MPRAGE','anat','anatomy'],
            'dti': ['DTI'],
            'field_map': ['fieldmap','TE7','B0']
        }
    '''
    min_acceptable_match = 80
    if image_fname:
        label_info = info_for_tags(image_fname,[(0x8,0x103e)])
        image_label = label_info[(0x8,0x103e)]
    # creates a list of tuples: (type, keyword)
    flat_dict = [i for j in [[(b,x) for x in label_dict[b]]  for b in label_dict] for i in j]
    best_match = process.extractOne(image_label,[x[1] for x in flat_dict])
    if best_match[1]<min_acceptable_match:
        return None
    else:
        return [x[0] for x in flat_dict if x[1]==best_match[0]][0]

def unpack_archive(fname,out_dir):
    '''unpacks the archive file ``fname`` and reconstructs datasets into ``out_dir``
    
    Datasets are reconstructed and auto-named using :meth:`create_dset`. The raw directories
    that made the datasets are archive with the dataset name suffixed by ``tgz``, and any other
    files found in the archive are put into ``other_files.tgz``'''
    with nl.notify('Unpacking archive %s' % fname):
        tmp_dir = tempfile.mkdtemp()
        tmp_unpack = os.path.join(tmp_dir,'unpack')
        os.makedirs(tmp_unpack)
        nl.utils.unarchive(fname,tmp_unpack)
        reconstruct_files(tmp_unpack)
        out_dir = os.path.abspath(out_dir)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        if not os.path.exists(tmp_unpack+'-sorted'):
            return
        with nl.run_in(tmp_unpack+'-sorted'):
            for fname in glob.glob('*.nii'):
                nl.run(['gzip',fname])
            for fname in glob.glob('*.nii.gz'):
                new_file = os.path.join(out_dir,fname)
                if not os.path.exists(new_file):
                    shutil.move(fname,new_file)
            raw_out = os.path.join(out_dir,'raw')
            if not os.path.exists(raw_out):
                os.makedirs(raw_out)
            for rawdir in os.listdir('.'):
                rawdir_tgz = os.path.join(raw_out,rawdir+'.tgz')
                if not os.path.exists(rawdir_tgz):
                    with tarfile.open(rawdir_tgz,'w:gz') as tgz:
                        tgz.add(rawdir)
        if len(os.listdir(tmp_unpack))!=0:
            # There are still raw files left
            with tarfile.open(os.path.join(raw_out,'other_files.tgz'),'w:gz') as tgz:
                tgz.add(tmp_unpack)
    shutil.rmtree(tmp_dir)


import sys,time,random
import zlib, base64

version_string = "Jarvis v0.1"

def compress(data, level=9):
	''' return compressed, ASCII-encoded string '''
	return base64.encodestring(zlib.compress(data,9))

def decompress(data):
	''' return uncompressed data '''
	return zlib.decompress(base64.decodestring(data))

def notify(msg):
	print msg

typing_speed = 100 #wpm
def slow_type(t):
	for l in t:
		sys.stdout.write(l)
		sys.stdout.flush()
		time.sleep(random.random()*10.0/typing_speed)
	print ''

try:
	from pyfiglet import Figlet
except ImportError:
	pass

def banner():
	try:
		return Figlet().renderText(version_string)
	except NameError:
		return version_string

def greet():
	notify('Hello there, sir.')

def goodbye():
	notify('All done.')


_unic_z = 'eNpVT8ENwzAI/DPFpR+SSIZnpczQDULqRRi+gGNVsSx8HBycCfMITaQ8oWoimcBN9U3ZgBeMsoAdO8GxZruViKUnk2+EwSBGarVw495aE2zYSPA4hzD5k4o1XnsqyO2UB/K6hQj2TcO98TGs50IfdQ5jMVqhNwGz1BhsDR/jp6mwqRgpPguWf4azX8Gd3ekHgEMzrA=='
_unic_bub_z = 'eNplUEEOgzAMu+cVZtIUmNTkOIk37Adk6yN2zeOXtFQTEFCp49Y2IYwSGjtloB5qUKpQnMqSlMY57l94YjfVZ+BLJRv6NxjyPRoppRAe8cQpBHTMzSPVwVKzk99Yss8Cph7JuHAtpQgWLCRH11WY/JLbm1tbZB8B951jZI+89sk0tfDaZ5CG3nmOXwjpiLo3YJZ3DDZHjjGy0Bg3OsRrwvRH2Oo7elt1+gFE3UfM'
_unic_bub_z_maxchars = 16

def _unic_bub_print(msg):
	if len(msg)>_unic_bub_z_maxchars:
		msg = msg[:_unic_bub_z_maxchars]
	msg = msg + ' '*(_unic_bub_z_maxchars - len(msg))
	msg = decompress(_unic_bub_z) % msg
	print msg		

if __name__ == "__main__":
	notify = _unic_bub_print
	greet()


''' wrapper functions for FSL programs '''
import neural as nl

fsl_dir = None
bet2 = None

def skull_strip(dset,suffix='_ns'):
    ''' use bet to strip skull from given anatomy '''
    # should add options to use betsurf and T1/T2 in the future
    out_dset = nl.afni.suffix(dset,suffix)
    cmd = bet2 if bet2 else 'bet2'
    cmd = os.path.join(fsl_dir,cmd) if fsl_dir else cmd
    nl.run([cmd,dset,out_dset],products=out_dset)


from distutils.core import setup
setup(
  name = 'neural-fmri',
  packages = ['neural'], # this must be the same as the name above
  version = '0.5',
  description = 'Neuroimaging Analysis Library',
  long_description = '''NeurAL - a neuroimaging analysis library
-------------------------------------------------------

**Documentation and examples:**
http://azraq27.github.com/neural

This library contains helper functions for doing analyses on fMRI data in Python.

In comparison to other Python libraries designed to interact with fMRI data
(e.g., NIPY and PyNIfTI), this library is not intended to interact directly with
the data in any way, just to provide helpful wrapper functions and shortcut methods to
make your life easier.

Since the author uses primarily AFNI, most of the functions are written that way, 
but don't specifically have to be that way...

''',
  author = 'Bill Gross',
  author_email = 'bill.gross@me.com',
  url = 'https://github.com/azraq27/neural',
  download_url = 'https://github.com/azraq27/neural/tarball/0.5',
  keywords = ['neuroimaging', 'afni', 'fsl', 'fmri'],
  classifiers = [
      'Topic :: Scientific/Engineering',
      'Intended Audience :: Science/Research',
      'Development Status :: 3 - Alpha'
  ],
)



''' generic imaging routines not specifically tied to any analysis packages '''

import nibabel as nib

class Analyze:
    ''' class to do arbitrary analyses on datasets
    
    Basically just a wrapper for NiBabel
    '''
    def __init__(self,dset=None):
        '''if ``dset`` is given, will automatically be loaded'''
        self.dset_filename = dset
        self.data = None
        self.header = None
        if dset:
            self.load(dset)
    
    def load(self,dset):
        '''load a dataset from given filename into the object'''
        self.dset_filename = dset
        self.dset = nib.load(dset)
        self.data = self.dset.get_data()
        self.header = self.dset.get_header()
    
    def voxel_loop(self):
        '''iterator that loops through each voxel and yields the coords and time series as a tuple'''
        # Prob not the most efficient, but the best I can do for now:
        for x in xrange(len(self.data)):
            for y in xrange(len(self.data[x])):
                for z in xrange(len(self.data[x][y])):
                    yield ((x,y,z),self.data[x][y][z])


# Color functions
colors = {
    'black':  90,
    'red':    91,
    'green':  92,
    'yellow': 93,
    'blue':   94,
    'magenta': 95,
    'cyan':   96,
    'white':  97 }

def color(mesg,col):
	return "\033[%sm%s\033[0m" %(colors[col],mesg)

def cursor_move_to(line,col):
	return "\033[%d;%dH" %(line,col)

def cursor_up(lines):
	return "\033[%dA" %(lines)

def cursor_down(lines):
	return "\033[%dB" %(lines)

def cursor_forward(cols):
	return "\033[%dC" %(cols)

def cursor_back(cols):
	return "\033[%dD" %(cols)

def cursor_clearscreen():
	return "\033[2J"

def cursor_clearline():
	return "\033[K"